from sigma.processing.conditions.base import (
    ProcessingCondition,
    DetectionItemProcessingCondition,
    FieldNameProcessingCondition,
    RuleProcessingCondition,
)
from typing import Mapping, Type

from sigma.processing.conditions.fields import ExcludeFieldCondition, IncludeFieldCondition
from sigma.processing.conditions.rule import (
    IsSigmaCorrelationRuleCondition,
    IsSigmaRuleCondition,
    LogsourceCondition,
    RuleAttributeCondition,
    RuleContainsDetectionItemCondition,
    RuleContainsFieldCondition,
    RuleTagCondition,
)
from sigma.processing.conditions.state import (
    DetectionItemProcessingItemAppliedCondition,
    DetectionItemProcessingStateCondition,
    FieldNameProcessingItemAppliedCondition,
    FieldNameProcessingStateCondition,
    RuleProcessingItemAppliedCondition,
    RuleProcessingStateCondition,
)
from sigma.processing.conditions.values import (
    ContainsWildcardCondition,
    IsNullCondition,
    MatchStringCondition,
    MatchValueCondition,
)


rule_conditions: Mapping[str, Type[RuleProcessingCondition]] = {
    "logsource": LogsourceCondition,
    "contains_detection_item": RuleContainsDetectionItemCondition,
    "contains_field": RuleContainsFieldCondition,
    "processing_item_applied": RuleProcessingItemAppliedCondition,
    "processing_state": RuleProcessingStateCondition,
    "is_sigma_rule": IsSigmaRuleCondition,
    "is_sigma_correlation_rule": IsSigmaCorrelationRuleCondition,
    "rule_attribute": RuleAttributeCondition,
    "tag": RuleTagCondition,
}
detection_item_conditions: Mapping[str, Type[DetectionItemProcessingCondition]] = {
    "match_string": MatchStringCondition,
    "match_value": MatchValueCondition,
    "contains_wildcard": ContainsWildcardCondition,
    "is_null": IsNullCondition,
    "processing_item_applied": DetectionItemProcessingItemAppliedCondition,
    "processing_state": DetectionItemProcessingStateCondition,
}
field_name_conditions: Mapping[str, Type[FieldNameProcessingCondition]] = {
    "include_fields": IncludeFieldCondition,
    "exclude_fields": ExcludeFieldCondition,
    "processing_item_applied": FieldNameProcessingItemAppliedCondition,
    "processing_state": FieldNameProcessingStateCondition,
}

__all__ = [
    "ProcessingCondition",
    "DetectionItemProcessingCondition",
    "FieldNameProcessingCondition",
    "RuleProcessingCondition",
    "LogsourceCondition",
    "RuleContainsDetectionItemCondition",
    "RuleContainsFieldCondition",
    "RuleProcessingItemAppliedCondition",
    "RuleProcessingStateCondition",
    "IsSigmaRuleCondition",
    "IsSigmaCorrelationRuleCondition",
    "RuleAttributeCondition",
    "RuleTagCondition",
    "MatchStringCondition",
    "MatchValueCondition",
    "ContainsWildcardCondition",
    "IsNullCondition",
    "DetectionItemProcessingItemAppliedCondition",
    "DetectionItemProcessingStateCondition",
    "IncludeFieldCondition",
    "ExcludeFieldCondition",
    "FieldNameProcessingItemAppliedCondition",
    "FieldNameProcessingStateCondition",
]



import sys
from .pipeline import dummy_test_pipeline, another_test_pipeline, YetAnotherTestPipeline

if "pytest" in sys.modules:
    __all__ = ["dummy_test_pipeline", "another_test_pipeline", "YetAnotherTestPipeline"]



from abc import abstractmethod
from typing import (
    Iterable,
    Optional,
    Union,
    Iterator,
)
from dataclasses import dataclass, field
from sigma.processing.transformations.base import StringValueTransformation, ValueTransformation
from sigma.exceptions import (
    SigmaValueError,
    SigmaConfigurationError,
)
from sigma.types import (
    Placeholder,
    SigmaString,
    SigmaType,
    SigmaRegularExpression,
    SpecialChars,
    SigmaQueryExpression,
)


@dataclass
class PlaceholderIncludeExcludeMixin:
    include: Optional[list[str]] = field(default=None)
    exclude: Optional[list[str]] = field(default=None)

    def check_exclusivity(self) -> None:
        if self.include is not None and self.exclude is not None:
            raise SigmaConfigurationError(
                "Placeholder transformation include and exclude lists can only be used exclusively!"
            )

    def is_handled_placeholder(self, p: Placeholder) -> bool:
        return (
            (self.include is None and self.exclude is None)
            or (self.include is not None and p.name in self.include)
            or (self.exclude is not None and p.name not in self.exclude)
        )


@dataclass
class BasePlaceholderTransformation(ValueTransformation, PlaceholderIncludeExcludeMixin):
    """
    Placeholder base transformation. The parameters include and exclude can contain variable names that
    are handled by this transformation. Unhandled placeholders are left as they are and must be handled by
    later transformations.
    """

    def __post_init__(self) -> None:
        self.check_exclusivity()
        return super().__post_init__()

    def apply_value(self, field: Optional[str], val: SigmaType) -> Union[
        None,
        SigmaString,
        Iterable[SigmaString],
        SigmaRegularExpression,
        Iterable[SigmaRegularExpression],
    ]:
        if isinstance(val, (SigmaString, SigmaRegularExpression)) and val.contains_placeholder(
            self.include, self.exclude
        ):
            return val.replace_placeholders(self.placeholder_replacements_base)
        else:
            return None

    def placeholder_replacements_base(
        self, p: Placeholder
    ) -> Iterator[Union[str, SpecialChars, Placeholder, SigmaString]]:
        """
        Base placeholder replacement callback. Calls real callback if placeholder is included or not excluded,
        else it passes the placeholder back to caller.
        """
        if self.is_handled_placeholder(p):
            yield from self.placeholder_replacements(p)
        else:
            yield p

    @abstractmethod
    def placeholder_replacements(
        self, p: Placeholder
    ) -> Iterable[Union[str, SpecialChars, Placeholder, SigmaString]]:
        """
        Placeholder replacement callback used by SigmaString.replace_placeholders(). This must return one
        of the following object types:

        * Plain strings
        * SpecialChars instances for insertion of wildcards
        * Placeholder instances, it may even return the same placeholder. These must be handled by following processing
          pipeline items or the backend or the conversion will fail.
        * SigmaString instances, these are used to replace the placeholder with a SigmaString that
          may contain plain strings, placeholders and special characters.
        """


@dataclass
class WildcardPlaceholderTransformation(BasePlaceholderTransformation):
    """
    Replaces placeholders with wildcards. This transformation is useful if remaining placeholders should
    be replaced with something meaningful to make conversion of rules possible without defining the
    placeholders content.
    """

    def placeholder_replacements(self, p: Placeholder) -> Iterable[SpecialChars]:
        return [SpecialChars.WILDCARD_MULTI]


@dataclass
class ValueListPlaceholderTransformation(BasePlaceholderTransformation):
    """
    Replaces placeholders with values contained in variables defined in the configuration.
    """

    def placeholder_replacements(self, p: Placeholder) -> Iterable[SigmaString]:
        try:
            if self._pipeline is None:
                raise SigmaValueError("No pipeline available for placeholder replacement.")
            values = self._pipeline.vars[p.name]
        except KeyError:
            raise SigmaValueError(f"Placeholder replacement variable '{ p.name }' doesn't exists.")

        if not isinstance(values, list):
            values = [values]

        if {isinstance(item, (str, int, float)) for item in values} != {True}:
            raise SigmaValueError(
                f"Replacement variable '{ p.name }' contains value which is not a string or number."
            )

        return [SigmaString(str(v)) for v in values]


@dataclass
class QueryExpressionPlaceholderTransformation(
    StringValueTransformation, PlaceholderIncludeExcludeMixin
):
    """
    Replaces a placeholder with a plain query containing the placeholder or an identifier
    mapped from the placeholder name. The main purpose is the generation of arbitrary
    list lookup expressions which are passed to the resulting query.

    Parameters:
    * expression: string that contains query expression with {field} and {id} placeholder
    where placeholder identifier or a mapped identifier is inserted.
    * mapping: Mapping between placeholders and identifiers that should be used in the expression.
    If no mapping is provided the placeholder name is used.
    """

    expression: str = ""
    mapping: dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.check_exclusivity()
        return super().__post_init__()

    def apply_string_value(self, field: Optional[str], val: SigmaString) -> Optional[SigmaType]:
        if val.contains_placeholder():
            if len(val.s) == 1 and isinstance(
                val.s[0], Placeholder
            ):  # Sigma string must only contain placeholder, nothing else.
                p = val.s[0]
                if self.is_handled_placeholder(p):
                    return SigmaQueryExpression(self.expression, self.mapping.get(p.name) or p.name)
            else:  # SigmaString contains placeholder as well as other parts
                raise SigmaValueError(
                    "Placeholder query expression transformation only allows placeholder-only strings."
                )
        return None



from typing import Any, Union
from dataclasses import dataclass
from sigma.correlations import SigmaCorrelationRule
from sigma.processing.transformations.base import PreprocessingTransformation
from sigma.rule import SigmaRule


@dataclass
class SetStateTransformation(PreprocessingTransformation):
    """Set pipeline state key to value."""

    key: str
    val: Any

    def apply(self, rule: Union[SigmaRule, SigmaCorrelationRule]) -> None:
        super().apply(rule)
        if self._pipeline is not None:
            self._pipeline.state[self.key] = self.val



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from enum import Enum, auto
from typing import ClassVar

from typing_extensions import Self

from sigma.correlations import SigmaCorrelationRule
from sigma.rule import SigmaDetection, SigmaDetectionItem, SigmaRule, SigmaRuleBase, SigmaRuleTag
from sigma.types import SigmaString, SigmaType


class SigmaValidationIssueSeverity(Enum):
    """
    Severity of a Sigma rule validation issue:

    * Low: minor improvement suggestion that results in better readability or maintainability of the
      rule.
    * Medium: issue can cause problems under certain conditions or the meaning of the rule can be
      different from intended.
    * High: issue will cause problems. It is certain that the intention of the rule author and the
      rule logic deviate or the rule doesn't match anything.
    """

    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()


@dataclass
class SigmaValidationIssue(ABC):
    """
    Describes an issue of one or multiple Sigma rules. This is a base class that should be
    overridden with specific issue classes. Description should contain some general issue
    information defined statically for the class. Additional issue information should be provided by
    additional fields that are automatically rendered in the representation methods.
    """

    description: ClassVar[str] = "Sigma rule validation issue"
    severity: ClassVar[SigmaValidationIssueSeverity]
    rules: list[SigmaRule | SigmaCorrelationRule]

    def __post_init__(self: Self) -> None:
        """Ensure that `self.rules` contains a list, even when a single rule was provided."""
        if isinstance(self.rules, SigmaRuleBase):
            self.rules = [self.rules]

    def __str__(self: Self) -> str:
        rules = ", ".join(
            [
                str(rule.source) if rule.source is not None else str(rule.id) or rule.title
                for rule in self.rules
            ]
        )
        additional_fields = " ".join(
            [
                f"{field.name}={self.__getattribute__(field.name) or '-'}"
                for field in fields(self)
                if field.name not in ("rules", "severity", "description")
            ]
        )
        return f'issue={self.__class__.__name__} severity={self.severity.name.lower()} description="{self.description}" rules=[{rules}] {additional_fields}'


class SigmaRuleValidator(ABC):
    """
    A rule validator class implements a check for a Sigma rule. It is instantiated once by
    SigmaCollectionValidator and can therefore keep a state across the validation of a whole Sigma
    collection. The validate() method returns results for a specific rule while finalize() is called
    at the end of the validation of multiple rules and can return issues that apply across multiple
    rules, e.g. violation of uniqueness constraints.
    """

    @abstractmethod
    def validate(self: Self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        """Implementation of the rule validation.

        :param rule: Sigma rule that should be validated.
        :type rule: SigmaRuleBase
        :return: List of validation issue objects describing.
        :rtype: list[SigmaValidationIssue]
        """
        self.rule = rule
        return []

    def finalize(self: Self) -> list[SigmaValidationIssue]:
        """
        Finalize a validation run and return validation issues that apply to multiple rules.

        :return: List of validation issues.
        :rtype: list[SigmaValidationIssue]
        """
        return []


class SigmaDetectionValidator(SigmaRuleValidator):
    """
    A detection validator class implements a check for detection definitions contained in Sigma
    rules. The method validate_detection() must be implemented and is called for each detection
    definition contained in the Sigma rule. It can perform isolated checks per detection or collect
    state across different detections and then conduct checks across multiple detections in the
    following methods:

    * validate(): all detections across a rule.
    * finalize(): all detections across a rule set.

    The validation state stored in the object should be reset as required to prevent undesired side
    effects in implementations of them methods mentioned above.
    """

    def validate(self: Self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        """
        Iterate over all detections and call validate_detection() for each.
        """
        super().validate(rule)
        if isinstance(rule, SigmaCorrelationRule):
            return []
        else:
            return [
                issue
                for name, detection in rule.detection.detections.items()
                for issue in self.validate_detection(name, detection)
            ]

    @abstractmethod
    def validate_detection(
        self: Self, name: str, detection: SigmaDetection
    ) -> list[SigmaValidationIssue]:
        """Implementation of the detection validation. It is invoked for each detection.

        :param name: Name of the validated detection.
        :type detection: str
        :param detection: detection definition that should be validated.
        :type detection: SigmaDetection
        :return: List of validation issue objects describing.
        :rtype: list[SigmaValidationIssue]
        """


class SigmaDetectionItemValidator(SigmaDetectionValidator):
    """
    A detection item validator iterates over all detection definitions and their detection items and
    calls the method validate_detection_item() for each of them. It can perform isolated checks per
    detection item or collect state across different detection items and then conduct checks across
    multiple of them in the following methods:

    * validate_detection(): all detection items of a detection.
    * validate(): all detection items across a rule.
    * finalize(): all detection items across a rule set.

    The validation state stored in the object should be reset as required to prevent undesired side
    effects in implementations of them methods mentioned above.
    """

    def validate_detection(
        self: Self, name: str | None, detection: SigmaDetection
    ) -> list[SigmaValidationIssue]:
        """
        Iterate over all detection items of a detection definition and call
        validate_detection_item() method on each detection item or this method itself recursively
        for nested detection definitions.
        """
        return [
            issue
            for item in detection.detection_items
            for issue in (
                self.validate_detection_item(item)
                if isinstance(item, SigmaDetectionItem)
                else self.validate_detection(None, item)
            )
        ]

    @abstractmethod
    def validate_detection_item(
        self: Self, detection_item: SigmaDetectionItem
    ) -> list[SigmaValidationIssue]:
        """Implementation of the detection item validation. It is invoked for each detection item.

        :param detection_item: detection item that should be validated.
        :type detection_item: SigmaDetectionItem
        :return: List of validation issue objects describing.
        :rtype: list[SigmaValidationIssue]
        """


class SigmaValueValidator(SigmaDetectionItemValidator):
    """
    A value validator iterates over all values contained in a Sigma rules detection items and calls
    the method validate_value() for each of them. It can perform isolated checks per value or collect state across different values and then
    conduct checks across multiple of them in the following methods:

    * validate_detection_item(): all values of a detection item.
    * validate_detection(): all detection items of a detection.
    * validate(): all detection items across a rule.
    * finalize(): all detection items across a rule set.

    The validation state stored in the object should be reset as required to prevent undesired side
    effects in implementations of them methods mentioned above.
    """

    def validate_detection_item(
        self: Self, detection_item: SigmaDetectionItem
    ) -> list[SigmaValidationIssue]:
        """
        Iterate over all values of a detection item and call validate_value() method for each of
        them.
        """
        return [issue for value in detection_item.value for issue in self.validate_value(value)]

    @abstractmethod
    def validate_value(self: Self, value: SigmaType) -> list[SigmaValidationIssue]:
        """Implementation of the value validation. It is invoked for each value of a type.

        :param value: detection item that should be validated.
        :type value: SigmaType
        :return: List of validation issue objects describing.
        :rtype: list[SigmaValidationIssue]
        """


class SigmaStringValueValidator(SigmaValueValidator):
    """
    A value validator iterates over all values contained in a Sigma rules detection items and calls
    the method validate_value() for all strings. It can perform isolated checks per value or collect
    state across different values and then conduct checks across multiple of them in the following
    methods:

    * validate_detection_item(): all values of a detection item.
    * validate_detection(): all detection items of a detection.
    * validate(): all detection items across a rule.
    * finalize(): all detection items across a rule set.

    The validation state stored in the object should be reset as required to prevent undesired side
    effects in implementations of them methods mentioned above.
    """

    def validate_value(self: Self, value: SigmaType) -> list[SigmaValidationIssue]:
        if isinstance(value, SigmaString):
            return self.validate_string(value)
        return []

    @abstractmethod
    def validate_string(self: Self, value: SigmaString) -> list[SigmaValidationIssue]:
        """Implementation of the string value validation. It is invoked for each SigmaString value.

        :param value: detection item that should be validated.
        :type value: SigmaString
        :return: List of validation issue objects describing.
        :rtype: list[SigmaValidationIssue]
        """


class SigmaTagValidator(SigmaRuleValidator):
    """
    The tag validator iterates over all tags from the rule and calls the method validate_tag() for
    each tag.
    """

    def validate(self: Self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        super().validate(rule)
        return [issue for tag in rule.tags for issue in self.validate_tag(tag)]

    @abstractmethod
    def validate_tag(self: Self, tag: SigmaRuleTag) -> list[SigmaValidationIssue]:
        """Validates a tag."""



from abc import ABC, abstractmethod
import re
from typing import (
    Any,
    ClassVar,
    Generator,
    Iterator,
    Optional,
    Union,
    Type,
    cast,
    get_origin,
    get_args,
    get_type_hints,
    Generic,
    TypeVar,
    TYPE_CHECKING,
)
from collections.abc import Sequence as SequenceABC
from base64 import b64encode
from sigma.types import (
    CompareOperators,
    Placeholder,
    SigmaBool,
    SigmaCasedString,
    SigmaExists,
    SigmaExpansion,
    SigmaFieldReference,
    SigmaRegularExpressionFlag,
    SigmaType,
    SigmaString,
    SigmaNumber,
    SpecialChars,
    SigmaRegularExpression,
    SigmaCompareExpression,
    SigmaCIDRExpression,
    SigmaTimestampPart,
    TimestampPart,
    SigmaStringPartType,
)
from sigma.exceptions import SigmaRuleLocation, SigmaTypeError, SigmaValueError

if TYPE_CHECKING:
    from .rule import SigmaDetectionItem

T = TypeVar("T", bound=Union[SigmaType, list[SigmaType]])
R = TypeVar("R", bound=Union[SigmaType, list[SigmaType]])


### Base Classes ###
class SigmaModifier(ABC, Generic[T, R]):
    """Base class for all Sigma modifiers"""

    detection_item: "SigmaDetectionItem"
    applied_modifiers: list[Type["SigmaModifier[T, R]"]]

    def __init__(
        self,
        detection_item: "SigmaDetectionItem",
        applied_modifiers: list[Type["SigmaModifier[T, R]"]],
        source: Optional[SigmaRuleLocation] = None,
    ):
        self.detection_item = detection_item
        self.applied_modifiers = applied_modifiers
        self.source = source

    def type_check(self, val: Any, explicit_type: Optional[Type[Any]] = None) -> bool:
        th = (
            explicit_type or get_type_hints(self.modify)["val"]
        )  # get type annotation from val parameter of apply method or explicit_type parameter
        if th is Any:
            return True
        to = get_origin(th)  # get possible generic type of type hint
        if to is None:  # Plain type in annotation
            return isinstance(val, th)
        elif to is Union:  # type hint is Union of multiple types, check if val is one of them
            for t in get_args(th):
                if isinstance(val, t):
                    return True
            return False
        elif to is list and isinstance(val, list):  # type hint is sequence
            inner_type = get_args(th)[0]
            return all([self.type_check(item, explicit_type=inner_type) for item in val])
        return False

    @abstractmethod
    def modify(self, val: T) -> R:
        """This method should be overridden with the modifier implementation."""

    def apply(self, val: T) -> list[T]:
        """
        Modifier entry point containing the default operations:
        * Type checking
        * Ensure returned value is a list
        * Handle values of SigmaExpansion objects separately.
        """
        if isinstance(val, SigmaExpansion):  # Handle each SigmaExpansion item separately
            return [
                cast(
                    T,
                    SigmaExpansion(
                        [cast(SigmaType, va) for v in val.values for va in self.apply(cast(T, v))]
                    ),
                )
            ]
        else:
            if not self.type_check(val):
                raise SigmaTypeError(
                    f"Modifier {self.__class__.__name__} incompatible to value type of '{ val }'",
                    source=self.source,
                )
            r = self.modify(val)
            if isinstance(r, list):
                return [cast(T, item) for item in r]
            else:
                return [cast(T, r)]


class SigmaValueModifier(SigmaModifier[T, R]):
    """Base class for all modifiers that handle each value for the modifier scope separately"""

    @abstractmethod
    def modify(self, val: T) -> R:
        """This method should be overridden with the modifier implementation."""


class SigmaListModifier(SigmaModifier[T, R]):
    """Base class for all modifiers that handle all values for the modifier scope as a whole."""

    @abstractmethod
    def modify(self, val: T) -> R:
        """This method should be overridden with the modifier implementation."""


### Modifier Implementations ###
class SigmaContainsModifier(
    SigmaValueModifier[
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
    ]
):
    """Puts wildcards around a string to match it somewhere inside another string instead of as a whole."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]
    ) -> Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]:
        if isinstance(val, SigmaString):
            if not val.startswith(SpecialChars.WILDCARD_MULTI):
                val = SpecialChars.WILDCARD_MULTI + val
            if not val.endswith(SpecialChars.WILDCARD_MULTI):
                val += SpecialChars.WILDCARD_MULTI
        elif isinstance(val, SigmaRegularExpression):
            regexp_str = str(val.regexp)
            if regexp_str[:2] != ".*" and regexp_str[0] != "^":
                val.regexp = SigmaString(".") + SpecialChars.WILDCARD_MULTI + val.regexp
            if regexp_str[-2:] != ".*" and regexp_str[-1] != "$":
                val.regexp += SigmaString(".") + SpecialChars.WILDCARD_MULTI
            val.compile()
        elif isinstance(val, SigmaFieldReference):
            val.starts_with = True
            val.ends_with = True
        return val


class SigmaStartswithModifier(
    SigmaValueModifier[
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
    ]
):
    """Puts a wildcard at the end of a string to match arbitrary values after the given prefix."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]
    ) -> Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]:
        if isinstance(val, SigmaString):
            if not val.endswith(SpecialChars.WILDCARD_MULTI):
                val += SpecialChars.WILDCARD_MULTI
        elif isinstance(val, SigmaRegularExpression):
            regexp_str = str(val.regexp)
            if regexp_str[-2:] != ".*" and regexp_str[-1] != "$":
                val.regexp += SigmaString(".") + SpecialChars.WILDCARD_MULTI
            val.compile()
        elif isinstance(val, SigmaFieldReference):
            val.starts_with = True
        return val


class SigmaEndswithModifier(
    SigmaValueModifier[
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
        Union[SigmaString, SigmaRegularExpression, SigmaFieldReference],
    ]
):
    """Puts a wildcard before a string to match arbitrary values before it."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]
    ) -> Union[SigmaString, SigmaRegularExpression, SigmaFieldReference]:
        if isinstance(val, SigmaString):
            if not val.startswith(SpecialChars.WILDCARD_MULTI):
                val = SpecialChars.WILDCARD_MULTI + val
        elif isinstance(val, SigmaRegularExpression):
            regexp_str = str(val.regexp)
            if regexp_str[:2] != ".*" and regexp_str[0] != "^":
                val.regexp = SigmaString(".") + SpecialChars.WILDCARD_MULTI + val.regexp
            val.compile()
        elif isinstance(val, SigmaFieldReference):
            val.ends_with = True
        return val


class SigmaBase64Modifier(SigmaValueModifier[SigmaString, SigmaString]):
    """Encode string as Base64 value."""

    def modify(self, val: SigmaString) -> SigmaString:
        if val.contains_special():
            raise SigmaValueError(
                "Base64 encoding of strings with wildcards is not allowed",
                source=self.source,
            )
        return SigmaString(b64encode(bytes(val)).decode())


class SigmaBase64OffsetModifier(SigmaValueModifier[SigmaString, SigmaExpansion]):
    """
    Encode string as Base64 value with different offsets to match it at different locations in
    encoded form.
    """

    start_offsets = (0, 2, 3)
    end_offsets = (None, -3, -2)

    def modify(self, val: SigmaString) -> SigmaExpansion:
        if val.contains_special():
            raise SigmaValueError(
                "Base64 encoding of strings with wildcards is not allowed",
                source=self.source,
            )
        return SigmaExpansion(
            [
                SigmaString(
                    b64encode(i * b" " + bytes(val))[
                        self.start_offsets[i] : self.end_offsets[(len(val) + i) % 3]
                    ].decode()
                )
                for i in range(3)
            ]
        )


class SigmaWideModifier(SigmaValueModifier[SigmaString, SigmaString]):
    """Encode string as wide string (UTF-16LE)."""

    def modify(self, val: SigmaString) -> SigmaString:
        r: list[SigmaStringPartType] = list()
        for item in val.s:
            if isinstance(
                item, str
            ):  # put 0x00 after each character by encoding it to utf-16le and decoding it as utf-8
                try:
                    r.append(item.encode("utf-16le").decode("utf-8"))
                except UnicodeDecodeError:  # this method only works for ascii characters
                    raise SigmaValueError(
                        f"Wide modifier only allowed for ascii strings, input string '{str(val)}' isn't one",
                        source=self.source,
                    )
            else:  # just append special characters without further handling
                r.append(item)

        s = SigmaString()
        s.s = r
        return s


class SigmaWindowsDashModifier(SigmaValueModifier[SigmaString, SigmaExpansion]):
    """
    Expand parameter characters / and - that are often interchangeable in Windows into the other
    form if it appears between word boundaries. E.g. in -param-name the first dash will be expanded
    into /param-name while the second dash is left untouched.
    """

    en_dash = chr(int("2013", 16))
    em_dash = chr(int("2014", 16))
    horizontal_bar = chr(int("2015", 16))

    def modify(self, val: SigmaString) -> SigmaExpansion:
        def callback(p: Placeholder) -> Iterator[Union[str, Placeholder]]:
            if p.name == "_windash":
                yield from ("-", "/", self.en_dash, self.em_dash, self.horizontal_bar)
            else:
                yield p

        return SigmaExpansion(
            cast(
                list[SigmaType],
                val.replace_with_placeholder(
                    re.compile("\\B[-/]\\b"), "_windash"
                ).replace_placeholders(callback),
            )
        )


class SigmaRegularExpressionModifier(SigmaValueModifier[SigmaString, SigmaRegularExpression]):
    """Treats string value as (case-sensitive) regular expression."""

    def modify(self, val: SigmaString) -> SigmaRegularExpression:
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "Regular expression modifier only applicable to unmodified values",
                source=self.source,
            )
        return SigmaRegularExpression(val.original)


class SigmaRegularExpressionFlagModifier(
    SigmaValueModifier[SigmaRegularExpression, SigmaRegularExpression]
):
    """Generic base class for setting a regular expression flag including checks"""

    flag: ClassVar[SigmaRegularExpressionFlag]

    def modify(self, val: SigmaRegularExpression) -> SigmaRegularExpression:
        val.add_flag(self.flag)
        return val


class SigmaRegularExpressionIgnoreCaseFlagModifier(SigmaRegularExpressionFlagModifier):
    """Match regular expression case-insensitive."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.IGNORECASE


class SigmaRegularExpressionMultilineFlagModifier(SigmaRegularExpressionFlagModifier):
    """Match regular expression across multiple lines."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.MULTILINE


class SigmaRegularExpressionDotAllFlagModifier(SigmaRegularExpressionFlagModifier):
    """Regular expression dot matches all characters."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.DOTALL


class SigmaCaseSensitiveModifier(SigmaValueModifier[SigmaString, SigmaCasedString]):
    def modify(self, val: SigmaString) -> SigmaCasedString:
        return SigmaCasedString.from_sigma_string(val)


class SigmaCIDRModifier(SigmaValueModifier[SigmaString, SigmaCIDRExpression]):
    """Treat value as IP (v4 or v6) CIDR network."""

    def modify(self, val: SigmaString) -> SigmaCIDRExpression:
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "CIDR expression modifier only applicable to unmodified values",
                source=self.source,
            )
        return SigmaCIDRExpression(str(val), source=self.source)


class SigmaAllModifier(SigmaListModifier[Any, Any]):
    """Match all values of a list instead of any pf them."""

    def modify(self, val: Any) -> Any:
        from sigma.conditions import ConditionAND

        self.detection_item.value_linking = ConditionAND
        return val


class SigmaCompareModifier(SigmaValueModifier[SigmaNumber, SigmaCompareExpression]):
    """Base class for numeric comparison operator modifiers."""

    op: ClassVar[CompareOperators]

    def modify(self, val: SigmaNumber) -> SigmaCompareExpression:
        return SigmaCompareExpression(val, self.op, self.source)


class SigmaLessThanModifier(SigmaCompareModifier):
    """Numeric less than (<) matching."""

    op: ClassVar[CompareOperators] = CompareOperators.LT


class SigmaLessThanEqualModifier(SigmaCompareModifier):
    """Numeric less than or equal (<=) matching."""

    op: ClassVar[CompareOperators] = CompareOperators.LTE


class SigmaGreaterThanModifier(SigmaCompareModifier):
    """Numeric greater than (>) matching."""

    op: ClassVar[CompareOperators] = CompareOperators.GT


class SigmaGreaterThanEqualModifier(SigmaCompareModifier):
    """Numeric greater than or equal (>=) matching."""

    op: ClassVar[CompareOperators] = CompareOperators.GTE


class SigmaNotEqualModifier(SigmaCompareModifier):
    """Numeric not equal (!=) matching."""

    op: ClassVar[CompareOperators] = CompareOperators.NEQ


class SigmaFieldReferenceModifier(SigmaValueModifier[SigmaString, SigmaFieldReference]):
    """Modifiers a plain string into the field reference type."""

    def modify(self, val: SigmaString) -> SigmaFieldReference:
        if val.contains_special():
            raise SigmaValueError("Field references must not contain wildcards", source=self.source)
        return SigmaFieldReference(val.to_plain())


class SigmaExistsModifier(SigmaValueModifier[SigmaBool, SigmaExists]):
    """Modifies to check if the field name provided as value exists in the matched event."""

    def modify(self, val: SigmaBool) -> SigmaExists:
        if self.detection_item.field is None:
            raise SigmaValueError("Exists modifier must be applied to field", source=self.source)
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "Exists modifier only applicable to unmodified boolean values",
                source=self.source,
            )
        return SigmaExists(val.boolean)


class SigmaExpandModifier(
    SigmaValueModifier[
        Union[SigmaString, SigmaRegularExpression], Union[SigmaString, SigmaRegularExpression]
    ]
):
    """
    Modifier for expansion of placeholders in values. It replaces placeholder strings (%something%)
    with stub objects that are later expanded to one or multiple strings or replaced with some SIEM
    specific list item or lookup by the processing pipeline.
    """

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression]
    ) -> Union[SigmaString, SigmaRegularExpression]:
        return val.insert_placeholders()


class SigmaTimestampModifier(SigmaValueModifier[SigmaNumber, SigmaTimestampPart]):
    """
    Base class for timestamp modifiers that parse the field as a datetime/timestamp and transform it to a specific part.
    """

    time_part_unit: ClassVar[TimestampPart]

    def modify(self, val: SigmaNumber) -> SigmaTimestampPart:
        return SigmaTimestampPart(self.time_part_unit, int(val.number))


class SigmaTimestampMinuteModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the minute number. Between 0 and 59.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.MINUTE


class SigmaTimestampHourModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the hour number. Between 0 and 23.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.HOUR


class SigmaTimestampDayModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the day of the month number. Between 1 and 31.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.DAY


class SigmaTimestampWeekModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the week of the year number. Between 1 and 52.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.WEEK


class SigmaTimestampMonthModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the month of the year number. Between 1 and 12.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.MONTH


class SigmaTimestampYearModifier(SigmaTimestampModifier):
    """
    Modifier that parses the field as a datetime/timestamp and transforms it to the year number.
    """

    time_part_unit: ClassVar[TimestampPart] = TimestampPart.YEAR


# Mapping from modifier identifier strings to modifier classes
modifier_mapping: dict[str, Type[SigmaModifier[Any, Any]]] = {
    "all": SigmaAllModifier,
    "base64": SigmaBase64Modifier,
    "base64offset": SigmaBase64OffsetModifier,
    "cased": SigmaCaseSensitiveModifier,
    "cidr": SigmaCIDRModifier,
    "contains": SigmaContainsModifier,
    "day": SigmaTimestampDayModifier,
    "dotall": SigmaRegularExpressionDotAllFlagModifier,
    "endswith": SigmaEndswithModifier,
    "exists": SigmaExistsModifier,
    "expand": SigmaExpandModifier,
    "fieldref": SigmaFieldReferenceModifier,
    "gt": SigmaGreaterThanModifier,
    "gte": SigmaGreaterThanEqualModifier,
    "hour": SigmaTimestampHourModifier,
    "i": SigmaRegularExpressionIgnoreCaseFlagModifier,
    "ignorecase": SigmaRegularExpressionIgnoreCaseFlagModifier,
    "lt": SigmaLessThanModifier,
    "lte": SigmaLessThanEqualModifier,
    "m": SigmaRegularExpressionMultilineFlagModifier,
    "minute": SigmaTimestampMinuteModifier,
    "month": SigmaTimestampMonthModifier,
    "multiline": SigmaRegularExpressionMultilineFlagModifier,
    "re": SigmaRegularExpressionModifier,
    "s": SigmaRegularExpressionDotAllFlagModifier,
    "startswith": SigmaStartswithModifier,
    "week": SigmaTimestampWeekModifier,
    "wide": SigmaWideModifier,
    "windash": SigmaWindowsDashModifier,
    "year": SigmaTimestampYearModifier,
}

# Mapping from modifier class to identifier
reverse_modifier_mapping: dict[str, str] = {
    modifier_class.__name__: identifier for identifier, modifier_class in modifier_mapping.items()
}



from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from sigma.exceptions import (
    SigmaPipelineNotAllowedForBackendError,
    SigmaPipelineNotFoundError,
)
from sigma.processing.pipeline import ProcessingPipeline
from typing import Iterable, Optional, Tuple, Union, cast, Callable
from collections import namedtuple


@dataclass
class ProcessingPipelineResolver:
    """
    A processing pipeline resolver resolves a list of pipeline specifiers into one summarized processing pipeline.
    It takes care of sorting by priority and resolution of filenames as well as pipeline name identifiers.
    """

    pipelines: dict[str, Union[ProcessingPipeline, Callable[[], ProcessingPipeline]]] = field(
        default_factory=dict
    )

    def add_pipeline_class(self, pipeline: ProcessingPipeline) -> None:
        """Add named processing pipeline object to resolver. This pipeline can be resolved by the name."""
        if pipeline.name is None:
            raise ValueError("Processing pipeline must be named to be resolvable.")
        self.pipelines[pipeline.name] = pipeline

    @classmethod
    def from_pipeline_list(
        cls, pipelines: Iterable[ProcessingPipeline]
    ) -> "ProcessingPipelineResolver":
        """Instantiate processing pipeline resolver from list of pipeline objects."""
        return cls({pipeline.name: pipeline for pipeline in pipelines if pipeline.name is not None})

    def list_pipelines(self) -> Iterable[tuple[str, ProcessingPipeline]]:
        """List identifier/processing pipeline tuples."""
        return ((id, self.resolve_pipeline(id)) for id in self.pipelines.keys())

    def resolve_pipeline(self, spec: str, target: Optional[str] = None) -> ProcessingPipeline:
        """
        Resolve single processing pipeline. It first tries to find a pipeline with this identifier
        in the registered pipelines. If this fails, *spec* is treated as file name. If this fails
        too, a *SigmaPipelineNotFoundError* is raised.

        If *target* is specified, an additional check of the compatibility of the specified backend
        to the resolved pipeline is conducted. A *SigmaPipelineNotAllowedForBackendError* is raised
        if this check fails.
        """
        try:
            pipeline = self.pipelines[spec]
            if callable(pipeline):
                resolved_pipeline = pipeline()
            else:
                resolved_pipeline = pipeline
            if target is not None and not (
                len(resolved_pipeline.allowed_backends) == 0
                or target in resolved_pipeline.allowed_backends
            ):
                raise SigmaPipelineNotAllowedForBackendError(spec, target)
            return resolved_pipeline
        except KeyError:  # identifier not found, try it as path
            try:
                with open(spec, "r") as f:
                    return ProcessingPipeline.from_yaml(f.read())
            except OSError as e:
                raise SigmaPipelineNotFoundError(spec)

    def resolve(
        self, pipeline_specs: list[str], target: Optional[str] = None
    ) -> ProcessingPipeline:
        """
        Resolve a list of

        * processing pipeline names from pipelines added to the resolver or
        * file paths containing processing pipeline YAML definitions or
        * directories containing processing pipelines YAML definitions

        into a consolidated processing pipeline.

        If *target* is specified this is passed in each *resolve_pipeline* call to perform a
        compatibility check for the usage of the specified backend with the pipeline.
        """

        PipelineInfo = namedtuple("PipelineInfo", ["pipeline", "priority", "path"])

        def resolve_path(spec: str) -> PipelineInfo:
            pipeline = self.resolve_pipeline(spec, target)
            return PipelineInfo(pipeline=pipeline, priority=pipeline.priority, path=spec)

        def resolve_spec(pipelines: list[PipelineInfo], spec: str) -> list[PipelineInfo]:
            spec_path = Path(spec.rstrip("/*"))
            if spec_path.is_dir():
                pipelines.extend([resolve_path(str(path)) for path in spec_path.glob("**/*.yml")])
            else:
                pipelines.append(resolve_path(spec))

            return pipelines

        pipelines: list[PipelineInfo] = reduce(resolve_spec, pipeline_specs, [])

        return (
            sum([p.pipeline for p in sorted(pipelines, key=lambda p: (p.priority, p.path))])
            or ProcessingPipeline()
        )



from __future__ import annotations

import datetime as dt
import re
from abc import abstractmethod
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import TYPE_CHECKING, Any
from uuid import UUID

import yaml
from typing_extensions import Self

import sigma.exceptions as sigma_exceptions
from sigma.rule.attributes import SigmaLevel, SigmaRelated, SigmaRuleTag, SigmaStatus

if TYPE_CHECKING:
    from sigma.conversion.state import ConversionState
    from sigma.exceptions import SigmaError, SigmaRuleLocation


class SigmaYAMLLoader(yaml.CSafeLoader):
    """Custom YAML loader implementing additional functionality for Sigma."""

    def construct_mapping(self, node: yaml.MappingNode, deep: bool = False) -> dict[Any, Any]:
        keys = set()
        for k, v in node.value:
            key = self.construct_object(k, deep=deep)  # type: ignore
            if key in keys:
                raise yaml.error.YAMLError("Duplicate key '{k}'")
            else:
                keys.add(key)

        return super().construct_mapping(node, deep)


@dataclass
class SigmaRuleBase:
    title: str = ""
    id: UUID | None = None
    name: str | None = None
    taxonomy: str = "sigma"
    related: SigmaRelated | None = None
    status: SigmaStatus | None = None
    description: str | None = None
    license: str | None = None
    references: list[str] = field(default_factory=list)
    tags: list[SigmaRuleTag] = field(default_factory=list)
    author: str | None = None
    date: dt.date | None = None
    modified: dt.date | None = None
    fields: list[str] = field(default_factory=list)
    falsepositives: list[str] = field(default_factory=list)
    level: SigmaLevel | None = None
    scope: list[str] | None = None

    errors: list[sigma_exceptions.SigmaError] = field(default_factory=list)
    source: SigmaRuleLocation | None = field(default=None, compare=False)
    custom_attributes: dict[str, Any] = field(compare=False, default_factory=dict)

    _backreferences: list[SigmaRuleBase] = field(
        init=False, default_factory=list, repr=False, compare=False
    )
    _conversion_result: list[Any] | None = field(
        init=False, default=None, repr=False, compare=False
    )
    _conversion_states: list[ConversionState] | None = field(
        init=False, default=None, repr=False, compare=False
    )
    _output: bool = field(init=False, default=True, repr=False, compare=False)

    def __post_init__(self: Self) -> None:
        for field in ("references", "tags", "fields", "falsepositives"):
            if self.__getattribute__(field) is None:
                self.__setattr__(field, [])
        if self.id is not None and not isinstance(
            self.id, UUID
        ):  # Try to convert rule id into UUID object, but keep it if not possible
            try:
                self.id = UUID(self.id)
            except ValueError:
                pass

    @classmethod
    def from_dict_common_params(
        cls: type[Self],
        rule: dict[str, Any],
        collect_errors: bool = False,
        source: SigmaRuleLocation | None = None,
    ) -> tuple[dict[str, Any], list[SigmaError]]:
        """
        Convert Sigma rule base parsed in dict structure into kwargs dict that can be passed to the
        class instantiation of an object derived from the SigmaRuleBase class and the errors list.
        This is intended to be called only by to_dict() methods for processing the general
        parameters defined in the base class.

        if collect_errors is set to False exceptions are collected in the errors property of the resulting
        SigmaRule object. Else the first recognized error is raised as exception.
        """
        errors = []

        def get_rule_as_date(name: str, exception_class: type[SigmaError]) -> date | None:
            """
            Accepted string based date formats are in range 1000-01-01 .. 3999-12-31:
              * XXXX-XX-XX                                 -- fully corresponds to yaml date format
              * XXXX/XX/XX, XXXX/XX/X, XXXX/X/XX, XXXX/X/X -- often occurs in the US-based sigmas
            Not accepted are ambiguous dates such as:
                2024-01-1, 24-1-24, 24/1/1, ...
            """
            nonlocal errors, rule, source
            value = rule.get(name)
            if (
                value is not None
                and not isinstance(value, date)
                and not isinstance(value, datetime)
            ):
                error = True
                try:
                    value = str(value)  # forcifully convert whatever the type is into string
                    accepted_regexps = (
                        "([1-3][0-9][0-9][0-9])-([01][0-9])-([0-3][0-9])",  # 1000-01-01 .. 3999-12-31
                        "([1-3][0-9][0-9][0-9])/([01]?[0-9])/([0-3]?[0-9])",  # 1000/1/1, 1000/01/01 .. 3999/12/31
                    )
                    for date_regexp in accepted_regexps:
                        matcher = re.fullmatch(date_regexp, value)
                        if matcher:
                            result = date(int(matcher[1]), int(matcher[2]), int(matcher[3]))
                            error = False
                            break
                except Exception:
                    pass
                if error:
                    errors.append(
                        exception_class(
                            f"Rule {name} '{ value }' is invalid, use yyyy-mm-dd", source=source
                        )
                    )
                    return None
                return result
            else:
                return value

        # Rule identifier may be empty or must be valid UUID
        rule_id = rule.get("id")
        if rule_id is not None:
            try:
                rule_id = UUID(rule_id)
            except ValueError:
                errors.append(
                    sigma_exceptions.SigmaIdentifierError(
                        "Sigma rule identifier must be an UUID", source=source
                    )
                )

        # Rule name
        rule_name = rule.get("name")
        if rule_name is not None:
            if not isinstance(rule_name, str):
                errors.append(
                    sigma_exceptions.SigmaTypeError(
                        "Sigma rule name must be a string", source=source
                    )
                )
            else:
                if rule_name == "":
                    errors.append(
                        sigma_exceptions.SigmaNameError(
                            "Sigma rule name must not be empty", source=source
                        )
                    )
                else:
                    rule_name = rule_name

        # Rule taxonomy
        rule_taxonomy = rule.get("taxonomy", "sigma")
        if rule_taxonomy is not None:
            if not isinstance(rule_taxonomy, str):
                errors.append(
                    sigma_exceptions.SigmaTaxonomyError(
                        "Sigma rule taxonomy must be a string", source=source
                    )
                )
            else:
                if rule_taxonomy == "":
                    errors.append(
                        sigma_exceptions.SigmaTaxonomyError(
                            "Sigma rule taxonomy must not be empty", source=source
                        )
                    )
                else:
                    rule_taxonomy = rule_taxonomy

        # Rule related validation
        rule_related = rule.get("related")
        if rule_related is not None:
            if not isinstance(rule_related, list):
                errors.append(
                    sigma_exceptions.SigmaRelatedError(
                        "Sigma rule related must be a list", source=source
                    )
                )
            else:
                try:
                    rule_related = SigmaRelated.from_dict(rule_related)
                except sigma_exceptions.SigmaRelatedError as e:
                    errors.append(e)

        # Rule level validation
        rule_level = rule.get("level")
        if rule_level is not None:
            try:
                rule_level = SigmaLevel[rule_level.upper()]
            except KeyError:
                errors.append(
                    sigma_exceptions.SigmaLevelError(
                        f"'{ rule_level }' is not a valid Sigma rule level", source=source
                    )
                )

        # Rule status validation
        rule_status = rule.get("status")
        if rule_status is not None:
            if not isinstance(rule_status, str):
                errors.append(
                    sigma_exceptions.SigmaStatusError(
                        "Sigma rule status cannot be a list", source=source
                    )
                )
            else:
                try:
                    rule_status = SigmaStatus[rule_status.upper()]
                except KeyError:
                    errors.append(
                        sigma_exceptions.SigmaStatusError(
                            f"'{ rule_status }' is not a valid Sigma rule status", source=source
                        )
                    )

        # parse rule date if existing
        rule_date = get_rule_as_date("date", sigma_exceptions.SigmaDateError)

        # parse rule modified if existing
        rule_modified = get_rule_as_date("modified", sigma_exceptions.SigmaModifiedError)

        # Rule fields validation
        rule_fields = rule.get("fields")
        if rule_fields is not None and not isinstance(rule_fields, list):
            errors.append(
                sigma_exceptions.SigmaFieldsError(
                    "Sigma rule fields must be a list",
                    source=source,
                )
            )

        # Rule falsepositives validation
        rule_falsepositives = rule.get("falsepositives")
        if rule_falsepositives is not None and not isinstance(rule_falsepositives, list):
            errors.append(
                sigma_exceptions.SigmaFalsePositivesError(
                    "Sigma rule falsepositives must be a list",
                    source=source,
                )
            )

        # Rule author validation
        rule_author = rule.get("author")
        if rule_author is not None and not isinstance(rule_author, str):
            errors.append(
                sigma_exceptions.SigmaAuthorError(
                    "Sigma rule author must be a string",
                    source=source,
                )
            )

        # Rule description validation
        rule_description = rule.get("description")
        if rule_description is not None and not isinstance(rule_description, str):
            errors.append(
                sigma_exceptions.SigmaDescriptionError(
                    "Sigma rule description must be a string",
                    source=source,
                )
            )

        # Rule references validation
        rule_references = rule.get("references")
        if rule_references is not None and not isinstance(rule_references, list):
            errors.append(
                sigma_exceptions.SigmaReferencesError(
                    "Sigma rule references must be a list",
                    source=source,
                )
            )

        # Rule title validation
        rule_title = rule.get("title")
        if rule_title is None:
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule must have a title",
                    source=source,
                )
            )
        elif not isinstance(rule_title, str):
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule title must be a string",
                    source=source,
                )
            )
        elif len(rule_title) > 256:
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule title length must not exceed 256 characters",
                    source=source,
                )
            )

        # Rule scope validation
        rule_scope = rule.get("scope")
        if rule_scope is not None and not isinstance(rule_scope, list):
            errors.append(
                sigma_exceptions.SigmaScopeError(
                    "Sigma rule scope must be a list",
                    source=source,
                )
            )
        # Rule license validation
        rule_license = rule.get("license")
        if rule_license is not None and not isinstance(rule_license, str):
            errors.append(
                sigma_exceptions.SigmaLicenseError(
                    "Sigma rule license must be a string",
                    source=source,
                )
            )

        if not collect_errors and errors:
            raise errors[0]

        return (
            {
                "title": rule_title,
                "id": rule_id,
                "name": rule_name,
                "taxonomy": rule_taxonomy,
                "related": rule_related,
                "level": rule_level,
                "status": rule_status,
                "description": rule_description,
                "references": rule_references,
                "tags": [SigmaRuleTag.from_str(tag) for tag in rule.get("tags", list())],
                "author": rule_author,
                "date": rule_date,
                "modified": rule_modified,
                "fields": rule_fields,
                "falsepositives": rule_falsepositives,
                "scope": rule_scope,
                "license": rule_license,
                "source": source,
                "custom_attributes": {
                    k: v
                    for k, v in rule.items()
                    if k
                    not in set(cls.__dataclass_fields__.keys())
                    - {"errors", "source", "applied_processing_items"}
                },
            },
            errors,
        )

    @classmethod
    @abstractmethod
    def from_dict(cls: type[Self], rule: dict[str, Any], collect_errors: bool = False) -> Self:
        """Convert dict input into SigmaRule object."""
        raise NotImplementedError(
            "from_dict method must be implemented in the derived class of SigmaRuleBase"
        )

    @classmethod
    def from_yaml(cls: type[Self], rule: str, collect_errors: bool = False) -> Self:
        """Convert YAML input string with single document into SigmaRule object."""
        parsed_rule = yaml.load(rule, SigmaYAMLLoader)
        return cls.from_dict(parsed_rule, collect_errors)

    def to_dict(self: Self) -> dict[str, Any]:
        """Convert rule object into dict."""
        d: dict[str, Any] = {
            "title": self.title,
        }
        # Convert to string where possible
        for field in ("id", "status", "level", "author", "description", "name"):
            if (s := self.__getattribute__(field)) is not None:
                d[field] = str(s)

        # copy list of strings
        for field in ("references", "fields", "falsepositives", "scope"):
            if (l := self.__getattribute__(field)) is not None and len(l) > 0:
                d[field] = l.copy()

        # the special cases
        if len(self.tags) > 0:
            d["tags"] = [str(tag) for tag in self.tags]
        if self.date is not None:
            d["date"] = self.date.isoformat()
        if self.modified is not None:
            d["modified"] = self.modified.isoformat()

        # custom attributes
        d.update(self.custom_attributes)

        return d

    def add_backreference(self: Self, rule: SigmaRuleBase) -> None:
        """Add backreference to another rule."""
        self._backreferences.append(rule)

    def referenced_by(self: Self, rule: SigmaRuleBase) -> bool:
        """Check if rule is referenced by another rule."""
        return rule in self._backreferences

    def set_conversion_result(self: Self, result: list[Any]) -> None:
        """Set conversion result."""
        self._conversion_result = result

    def get_conversion_result(self: Self) -> list[Any]:
        """Get conversion result."""
        if self._conversion_result is None:
            raise sigma_exceptions.SigmaConversionError(
                self,
                None,
                "Conversion result not available",
            )
        return self._conversion_result

    def set_conversion_states(self: Self, state: list[ConversionState]) -> None:
        """Set conversion state."""
        self._conversion_states = state

    def get_conversion_states(self: Self) -> list[ConversionState]:
        """Get conversion state."""
        if self._conversion_states is None:
            raise sigma_exceptions.SigmaConversionError(
                self,
                None,
                "Conversion state not available",
            )
        return self._conversion_states

    def disable_output(self: Self) -> None:
        """Disable output of rule."""
        self._output = False

    def __lt__(self: Self, other: SigmaRuleBase) -> bool:
        """Sort rules by backreference. A rule referenced by another rule is smaller."""
        return self.referenced_by(other)



from __future__ import annotations

from collections import Counter, defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar
from uuid import UUID

from sigma.validators.base import (
    SigmaRuleValidator,
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
)

if TYPE_CHECKING:
    from sigma.correlations import SigmaCorrelationRule
    from sigma.rule import SigmaRule


def is_uuid_v4(val: str) -> bool:
    try:
        id = UUID(str(val))
        if id.version == 4:
            return True
        else:
            return False
    except ValueError:
        return False


@dataclass
class IdentifierExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no identifier (UUID)"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class IdentifierExistenceValidator(SigmaRuleValidator):
    """Checks if rule has identifier."""

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.id is None:
            return [IdentifierExistenceIssue([rule])]
        else:
            return []


@dataclass
class IdentifierCollisionIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule identifier used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    identifier: UUID


class IdentifierUniquenessValidator(SigmaRuleValidator):
    """Check rule UUID uniqueness."""

    ids: dict[UUID, list[SigmaRule | SigmaCorrelationRule]]

    def __init__(self) -> None:
        self.ids = defaultdict(list)

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.id is not None:
            self.ids[rule.id].append(rule)
        return []

    def finalize(self) -> list[SigmaValidationIssue]:
        return [
            IdentifierCollisionIssue(rules, id) for id, rules in self.ids.items() if len(rules) > 1
        ]


@dataclass
class DuplicateTitleIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule title used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    title: str


class DuplicateTitleValidator(SigmaRuleValidator):
    """Check rule title uniqueness."""

    titles: dict[str, list[SigmaRule | SigmaCorrelationRule]]

    def __init__(self) -> None:
        self.titles = defaultdict(list)

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.title is not None:
            self.titles[rule.title].append(rule)
        return []

    def finalize(self) -> list[SigmaValidationIssue]:
        return [
            DuplicateTitleIssue(rules, title)
            for title, rules in self.titles.items()
            if len(rules) > 1
        ]


@dataclass
class DuplicateReferencesIssue(SigmaValidationIssue):
    description: ClassVar[str] = "The same references appears multiple times"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    reference: str


class DuplicateReferencesValidator(SigmaRuleValidator):
    """Validate rule References uniqueness."""

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        references = Counter(rule.references)
        return [
            DuplicateReferencesIssue([rule], reference)
            for reference, count in references.items()
            if count > 1
        ]


@dataclass
class DuplicateFilenameIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule filename used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    filename: str


class DuplicateFilenameValidator(SigmaRuleValidator):
    """Check rule filename uniqueness."""

    filenames_to_rules: dict[str, list[SigmaRule | SigmaCorrelationRule]]
    filenames_to_paths: dict[str, set[str]]

    def __init__(self) -> None:
        self.filenames_to_rules = defaultdict(list)
        self.filenames_to_paths = defaultdict(set)

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.source is not None:
            self.filenames_to_rules[rule.source.path.name].append(rule)
            self.filenames_to_paths[rule.source.path.name].add(str(rule.source.path))
        return []

    def finalize(self) -> list[SigmaValidationIssue]:
        return [
            DuplicateFilenameIssue(self.filenames_to_rules[filename], filename)
            for filename, paths in self.filenames_to_paths.items()
            if len(paths) > 1
        ]


@dataclass
class FilenameLengthIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule filename is too short or long"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    filename: str


@dataclass(frozen=True)
class FilenameLengthValidator(SigmaRuleValidator):
    """Check rule filename length"""

    min_size: int = 10
    max_size: int = 90

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.source is not None:
            filename = rule.source.path.name
            if len(filename) < self.min_size or len(filename) > self.max_size:
                return [FilenameLengthIssue([rule], filename)]
        return []


@dataclass
class CustomAttributesIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule use optional field name similar to legit"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    fieldname: str


class CustomAttributesValidator(SigmaRuleValidator):
    """Check if field name is similar to legit one"""

    known_custom_attributes: ClassVar[set[str]] = {
        "realted",
        "relatde",
        "relted",
        "rlated",
        "reference",
    }

    def validate(self, rule: SigmaRule | SigmaCorrelationRule) -> list[SigmaValidationIssue]:
        if rule.custom_attributes is not None:
            for k in rule.custom_attributes.keys():
                if k in self.known_custom_attributes:
                    return [CustomAttributesIssue([rule], k)]
        return []



from dataclasses import dataclass, field
from typing import ClassVar, Union
from sigma.correlations import SigmaCorrelationRule
from sigma.rule import SigmaDetectionItem, SigmaRule
from sigma.types import SigmaNumber

from sigma.validators.base import (
    SigmaDetectionItemValidator,
    SigmaRuleValidator,
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
)
from sigma.rule import SigmaLogSource

specific_to_generic_logsource_mapping: dict[SigmaLogSource, dict[int, str]] = {
    SigmaLogSource(None, "windows", "sysmon"): {
        1: "process_creation",
        2: "file_change",
        3: "network_connection",
        5: "process_termination",
        4: "sysmon_status",
        6: "driver_load",
        7: "image_load",
        8: "create_remote_thread",
        9: "raw_access_thread",
        10: "process_access",
        11: "file_event",
        12: "registry_add",
        12: "registry_delete",
        13: "registry_set",
        14: "registry_rename",
        12: "registry_event",
        13: "registry_event",
        14: "registry_event",
        15: "create_stream_hash",
        16: "sysmon_status",
        17: "pipe_created",
        18: "pipe_created",
        19: "wmi_event",
        20: "wmi_event",
        21: "wmi_event",
        22: "dns_query",
        23: "file_delete",
        24: "clipboard_change",
        25: "process_tampering",
        26: "file_delete_detected",
        27: "file_block_executable",
        28: "file_block_shredding",
        29: "file_executable_detected",
        255: "sysmon_error",
    },
    SigmaLogSource(None, "windows", "security"): {
        4688: "process_creation",
    },
}


@dataclass
class SpecificInsteadOfGenericLogsourceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Usage of specific instead of generic log source"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    logsource: SigmaLogSource
    event_id: int
    generic_logsource: SigmaLogSource


class SpecificInsteadOfGenericLogsourceValidator(SigmaDetectionItemValidator):
    """Identify usage of specific Windows event identifiers where corresponding generic log sources exist."""

    def validate(self, rule: Union[SigmaRule, SigmaCorrelationRule]) -> list[SigmaValidationIssue]:
        if isinstance(rule, SigmaCorrelationRule):
            return []  # Correlation rules do not have detections

        for (
            logsource,
            eventid_mappings,
        ) in specific_to_generic_logsource_mapping.items():
            if rule.logsource in logsource:
                self.logsource = logsource
                self.eventid_mappings = eventid_mappings
                self.disallowed_logsource_event_ids = eventid_mappings.keys()
                return super().validate(rule)
        return []

    def validate_detection_item(
        self, detection_item: SigmaDetectionItem
    ) -> list[SigmaValidationIssue]:
        if detection_item.field == "EventID":
            return [
                SpecificInsteadOfGenericLogsourceIssue(
                    rules=[self.rule],
                    logsource=self.logsource,
                    event_id=int(event_id.number),
                    generic_logsource=SigmaLogSource(self.eventid_mappings[int(event_id.number)]),
                )
                for event_id in detection_item.value
                if isinstance(event_id, SigmaNumber)
                and event_id.number in self.disallowed_logsource_event_ids
            ]
        else:
            return []


@dataclass
class FieldnameLogsourceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Usage of invalid field name in the log source"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    fieldname: str


class FieldnameLogsourceValidator(SigmaRuleValidator):
    """Identify usage of invalid field names in the log source."""

    def validate(self, rule: Union[SigmaRule, SigmaCorrelationRule]) -> list[SigmaValidationIssue]:
        if isinstance(rule, SigmaRule) and rule.logsource.custom_attributes:
            return [
                FieldnameLogsourceIssue(rules=[rule], fieldname=name)
                for name in rule.logsource.custom_attributes
            ]
        else:
            return []



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from typing_extensions import Self

import sigma.exceptions as sigma_exceptions
from sigma.exceptions import SigmaError, SigmaRuleLocation
from sigma.processing.tracking import ProcessingItemTrackingMixin
from sigma.rule.base import SigmaRuleBase
from sigma.rule.detection import EmptySigmaDetections, SigmaDetections
from sigma.rule.logsource import EmptyLogSource, SigmaLogSource


@dataclass
class SigmaRule(SigmaRuleBase, ProcessingItemTrackingMixin):
    """
    A single Sigma rule.
    """

    logsource: SigmaLogSource = field(default_factory=SigmaLogSource)
    detection: SigmaDetections = field(default_factory=EmptySigmaDetections)

    @classmethod
    def from_dict(
        cls: type[Self],
        rule: dict[str, Any],
        collect_errors: bool = False,
        source: SigmaRuleLocation | None = None,
    ) -> Self:
        """
        Convert Sigma rule parsed in dict structure into SigmaRule object.

        if collect_errors is set to False exceptions are collected in the errors property of the resulting
        SigmaRule object. Else the first recognized error is raised as exception.
        """
        kwargs, errors = super().from_dict_common_params(rule, collect_errors, source)

        # parse log source
        try:
            logsource = SigmaLogSource.from_dict(rule["logsource"], source)
        except KeyError:
            logsource = EmptyLogSource()
            errors.append(
                sigma_exceptions.SigmaLogsourceError(
                    "Sigma rule must have a log source", source=source
                )
            )
        except AttributeError:
            logsource = EmptyLogSource()
            errors.append(
                sigma_exceptions.SigmaLogsourceError(
                    "Sigma logsource must be a valid YAML map", source=source
                )
            )
        except SigmaError as e:
            logsource = EmptyLogSource()
            errors.append(e)

        # parse detections
        try:
            detections = SigmaDetections.from_dict(rule["detection"], source)
        except KeyError:
            detections = EmptySigmaDetections()
            errors.append(
                sigma_exceptions.SigmaDetectionError(
                    "Sigma rule must have a detection definitions", source=source
                )
            )
        except SigmaError as e:
            detections = EmptySigmaDetections()
            errors.append(e)

        if not collect_errors and errors:
            raise errors[0]

        return cls(
            logsource=logsource,
            detection=detections,
            errors=errors,
            **kwargs,
        )

    @classmethod
    def from_yaml(cls: type[Self], rule: str, collect_errors: bool = False) -> Self:
        """Convert YAML input string with single document into SigmaRule object."""
        return super().from_yaml(rule, collect_errors)

    def to_dict(self) -> dict[str, Any]:
        """Convert rule object into dict."""
        d = super().to_dict()
        d.update(
            {
                "logsource": self.logsource.to_dict(),
                "detection": self.detection.to_dict(),
            }
        )

        return d



from dataclasses import dataclass, field
from datetime import date
from uuid import UUID

from sigma.correlations import SigmaCorrelationRule
from sigma.processing.conditions.base import (
    RuleDetectionItemCondition,
    RuleProcessingCondition,
)
from sigma.types import sigma_type
from typing import ClassVar, Literal, Optional, Union
from sigma.rule import (
    SigmaDetection,
    SigmaLevel,
    SigmaRule,
    SigmaDetectionItem,
    SigmaLogSource,
    SigmaRuleTag,
    SigmaStatus,
)
from sigma.exceptions import SigmaConfigurationError


@dataclass
class LogsourceCondition(RuleProcessingCondition):
    """
    Matches log source on rule. Not specified log source fields are ignored. For Correlation rules,
    the condition returns true if any of the associated rules have the required log source fields.
    """

    category: Optional[str] = field(default=None)
    product: Optional[str] = field(default=None)
    service: Optional[str] = field(default=None)

    def __post_init__(self) -> None:
        self.logsource = SigmaLogSource(self.category, self.product, self.service)

    def match(
        self,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> bool:
        if isinstance(rule, SigmaRule):
            return rule.logsource in self.logsource
        elif isinstance(rule, SigmaCorrelationRule):
            # Will only return true if the rules have been resolved in advance
            for ref in rule.referenced_rules:
                if hasattr(ref, "rule") and isinstance(ref.rule, (SigmaRule, SigmaCorrelationRule)):
                    if self.match(ref.rule):
                        return True
            return False


@dataclass
class RuleContainsFieldCondition(RuleDetectionItemCondition):
    """Returns True if rule contains a field that matches the given field name."""

    field: Optional[str]

    def find_detection_item(self, detection: Union[SigmaDetectionItem, SigmaDetection]) -> bool:
        if isinstance(detection, SigmaDetection):
            for detection_item in detection.detection_items:
                if self.find_detection_item(detection_item):
                    return True
        elif isinstance(detection, SigmaDetectionItem):
            if detection.field is not None and detection.field == self.field:
                return True
        else:
            raise TypeError("Parameter of type SigmaDetection or SigmaDetectionItem expected.")

        return False


@dataclass
class RuleContainsDetectionItemCondition(RuleDetectionItemCondition):
    """Returns True if rule contains a detection item that matches the given field name and value."""

    field: Optional[str]
    value: Union[str, int, float, bool]

    def __post_init__(self) -> None:
        self.sigma_value = sigma_type(self.value)

    def find_detection_item(self, detection: Union[SigmaDetectionItem, SigmaDetection]) -> bool:
        if isinstance(detection, SigmaDetection):
            for detection_item in detection.detection_items:
                if self.find_detection_item(detection_item):
                    return True
        elif isinstance(detection, SigmaDetectionItem):
            if (
                detection.field is not None
                and detection.field == self.field
                and self.sigma_value
                in [v for v in detection.value if isinstance(self.sigma_value, type(v))]
            ):
                return True
        else:
            raise TypeError("Parameter of type SigmaDetection or SigmaDetectionItem expected.")

        return False


@dataclass
class IsSigmaRuleCondition(RuleProcessingCondition):
    """
    Checks if rule is a SigmaRule.
    """

    def match(
        self,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> bool:
        return isinstance(rule, SigmaRule)


@dataclass
class IsSigmaCorrelationRuleCondition(RuleProcessingCondition):
    """
    Checks if rule is a SigmaCorrelationRule.
    """

    def match(
        self,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> bool:
        return isinstance(rule, SigmaCorrelationRule)


@dataclass
class RuleAttributeCondition(RuleProcessingCondition):
    """
    Generic match on rule attributes with supported types:

    * strings (exact matches with eq/ne)
    * UUIDs (exact matches with eq/ne)
    * lists (membership checks with in/not_in; eq/gte/gt/lte/lt always return False, ne always returns True)
    * numbers (relations: eq, ne, gte, ge, lte, le)
    * dates (relations: eq, ne, gte, ge, lte, le)
    * Rule severity levels (relations: eq, ne, gte, ge, lte, le)
    * Rule statuses (relations: eq, ne, gte, ge, lte, le)

    Fields that contain maps or other complex data structures are not supported and
    raise a SigmaConfigurationError. If the type of the value doesn't allows a particular relation, the
    condition also raises a SigmaConfigurationError on match.
    """

    attribute: str
    value: Union[str, int, float]
    op: Literal["eq", "ne", "gte", "gt", "lte", "lt", "in", "not_in"] = field(default="eq")
    op_methods: ClassVar[dict[str, str]] = {
        "eq": "__eq__",
        "ne": "__ne__",
        "gte": "__ge__",
        "gt": "__gt__",
        "lte": "__le__",
        "lt": "__lt__",
    }

    def __post_init__(self) -> None:
        if self.op not in self.op_methods and self.op not in ("in", "not_in"):
            raise SigmaConfigurationError(
                f"Invalid operation '{self.op}' in rule attribute condition {str(self)}."
            )

    def match(
        self,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> bool:
        try:  # first try to get built-in attribute
            value = getattr(rule, self.attribute)
        except AttributeError:
            try:
                value = rule.custom_attributes[self.attribute]
            except KeyError:
                return False

        # Finally, value has some comparable type
        compare_value: Union[str, int, float, date, SigmaLevel, SigmaStatus]
        if isinstance(value, list):  # list membership check
            if self.op == "in":
                return self.value in value
            elif self.op == "not_in":
                return self.value not in value
            elif self.op == "eq":
                return False  # A list will never equal a single value
            elif self.op == "ne":
                return True  # A list will never equal a single value
            elif self.op in ("gte", "gt", "lte", "lt"):
                return False  # A list cannot be compared numerically to a single value
            else:
                raise SigmaConfigurationError(
                    f"Invalid operation '{self.op}' for list comparison in rule attribute condition {str(self)}."
                )
        elif isinstance(value, (str, UUID)):  # exact match of strings and UUIDs
            if self.op == "eq":
                return str(value) == self.value
            elif self.op == "ne":
                return str(value) != self.value
            else:
                raise SigmaConfigurationError(
                    f"Invalid operation '{self.op}' for string comparison in rule attribute condition {str(self)}."
                )
        elif isinstance(value, (int, float)):  # numeric comparison
            try:
                compare_value = float(self.value)
            except ValueError:
                raise SigmaConfigurationError(
                    f"Invalid number format '{self.value}' in rule attribute condition {str(self)}."
                )
        elif isinstance(value, date):  # date comparison
            if not isinstance(self.value, str):
                raise SigmaConfigurationError(
                    f"'{self.value}' must be a string value with a valid date but is a {type(self.value)} in rule attribute condition {str(self)}."
                )
            try:
                compare_value = date.fromisoformat(self.value)
            except ValueError:
                raise SigmaConfigurationError(
                    f"Invalid date format '{self.value}' in rule attribute condition {str(self)}."
                )
        elif isinstance(value, SigmaLevel):
            try:
                if not isinstance(self.value, str):
                    raise SigmaConfigurationError(
                        f"'{self.value}' must be a string value with a valid severity level but is a {type(self.value)} in rule attribute condition {str(self)}."
                    )
                compare_value = SigmaLevel[self.value.upper()]
            except KeyError:
                raise SigmaConfigurationError(
                    f"Invalid Sigma severity level '{self.value}' in rule attribute condition {str(self)}."
                )
        elif isinstance(value, SigmaStatus):
            try:
                if not isinstance(self.value, str):
                    raise SigmaConfigurationError(
                        f"'{self.value}' must be a string value with a valid status but is a {type(self.value)} in rule attribute condition {str(self)}."
                    )
                compare_value = SigmaStatus[self.value.upper()]
            except KeyError:
                raise SigmaConfigurationError(
                    f"Invalid Sigma status '{self.value}' in rule attribute condition {str(self)}."
                )
        else:
            raise SigmaConfigurationError(
                f"Unsupported type '{type(value)}' in rule attribute condition {str(self)}."
            )

        try:
            return bool(getattr(value, self.op_methods[self.op])(compare_value))
        # bool(NotImplemented) used to return `True` with Python<3.14
        except TypeError:
            return True
        except AttributeError:  # operation not supported by value type
            return False


@dataclass
class RuleTagCondition(RuleProcessingCondition):
    """
    Matches if rule is tagged with a specific tag.
    """

    tag: str

    def __post_init__(self) -> None:
        self.match_tag = SigmaRuleTag.from_str(self.tag)

    def match(
        self,
        rule: Union[SigmaRule, SigmaCorrelationRule],
    ) -> bool:
        return self.match_tag in rule.tags
